#include "scripts.h"
#include "engine.h"
#include "Misc.h"
extern "C" 
{
	#include "LuaLib\lua.h"
	#include "LuaLib\lualib.h"
	#include "LuaLib\lauxlib.h"
}
#include <Windows.h>
#include <Winsock2.h>
#include "LuaManager.h"
#include "LuaFunctions.h"
#define new new(__FILE__, __LINE__)

#pragma warning(disable: 4291)
AutoStack<LuaScriptManager::Script *> LuaScriptManager::Scripts = AutoStack<LuaScriptManager::Script *>();


ScriptImpClass *LuaScriptManager::Get_Script(GameObject *obj, int scriptID)
{
	SimpleDynVecClass<GameObjObserverClass *> *observers = &((ScriptableGameObj *)obj)->Observers;
	int x = observers->Count();
	for (int i = 0;i < x;i++)
	{
		if((*observers)[i]->GetID() == scriptID)
		{
			return ((ScriptImpClass*)(*observers)[i]);
		}
	}
	return 0;
}



void LuaScriptManager::Lua_DestroyScript(GameObject *obj, int scriptID)
{
	if(!obj)
	{
		return;
	}
	ScriptImpClass *script = Get_Script(obj, scriptID);
	if(!script)
	{
		return;
	}
	script->Destroy_Script();
}

float LuaScriptManager::Lua_Get_Float_Parameter(GameObject *obj, int scriptID, const char *Name)
{
	if(!obj)
	{
		return 0;
	}
	ScriptImpClass *script = Get_Script(obj, scriptID);
	if(!script)
	{
		return 0;
	}
	return script->Get_Float_Parameter(Name);
}

int LuaScriptManager::Lua_Get_Int_Parameter(GameObject *obj, int scriptID, const char *Name)
{
	if(!obj)
	{
		return 0;
	}
	ScriptImpClass *script = Get_Script(obj, scriptID);
	if(!script)
	{
		return 0;
	}
	return script->Get_Int_Parameter(Name);
}

const char *LuaScriptManager::Lua_Get_String_Parameter(GameObject *obj, int scriptID, const char *Name)
{
	if(!obj)
	{
		return "";
	}
	ScriptImpClass *script = Get_Script(obj, scriptID);
	if(!script)
	{
		return "";
	}
	return script->Get_Parameter(Name);
}

void LuaScriptManager::AddScript(const char *LuaFile)
{
	lua_State *L = lua_open();

	luaopen_io(L); 
	luaopen_base(L);
	luaopen_table(L);
	luaopen_string(L);
	luaopen_math(L);
	luaopen_loadlib(L);
	AddFunctions(L);
	RegisterScriptFunctions(L);
	LuaSocketManager::Register(L);
	int s = luaL_loadfile(L, LuaFile);
	if (s == 0) 
	{
		s = lua_pcall(L, 0, LUA_MULTRET, 0);
	}
	if(Report_Errors(L, s))
	{
		lua_close(L);
		return;
	}
	const char *Name, *Params;
	
	lua_getglobal(L, "ScriptName");
	if(lua_isfunction(L, -1))
	{
		lua_call(L, 0, 1);
		Name = lua_tostring(L, -1);
		lua_pop(L, 1);
	}
	else
	{
		printf("Failed to load script inside %s: could not find function \"ScriptParams\"\n", LuaFile);
		lua_close(L);
		return;
	}
		
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		if(strcmp((*Scripts.Node())->script, Name) == 0)
		{
			printf("Found duplicate script: \"%s\", not adding script.", Name);
			lua_close(L);
			return;
		}
	}
	


	lua_getglobal(L, "ScriptParams");
	if(lua_isfunction(L, -1))
	{
		lua_call(L, 0, 1);
		Params = lua_tostring(L, -1);
		lua_pop(L, 1);
	}
	else
	{
		printf("Failed to load script inside %s: could not find function \"ScriptParams\"\n", LuaFile);
		lua_close(L);
		return;
	}

	Script *script = new Script;
	ScriptRegistrant<LuaScript> *reg = new ScriptRegistrant<LuaScript>((char *)Name, (char *)Params);
	script->script = reg->GetName();
	script->lua = L;
	script->scriptreg = reg;
	Scripts.Push(script);
	printf("Loaded script: %s\n", Name);
	
}

bool LuaScriptManager::Report_Errors(lua_State *L, int status)
{
	if(status != 0) 
	{
		printf("%s\n", lua_tostring(L, -1));
		lua_pop(L, 1); 
		return 1;
	}
	return 0;
}

void LuaScriptManager::Lua_Start_Timer(GameObject *obj, int scriptID, float Time, int Number)
{
	if(!obj)
	{
		return;
	}
	ScriptImpClass *script = Get_Script(obj, scriptID);
	if(!script)
	{
		return;
	}
	Commands->Start_Timer(obj, script, Time, Number);
}



void LuaScriptManager::Cleanup()
{
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		lua_close((*Scripts.Node())->lua);
		delete (*Scripts.Node())->scriptreg;
		delete (*Scripts.Node());
	}
}

void LuaScriptManager::RegisterScriptFunctions(lua_State *L)
{
	lua_register(L, "DestroyScript", LuaScriptManager::Lua_DestroyScript_Wrap);
	lua_register(L, "Get_Float_Parameter", LuaScriptManager::Lua_Get_Float_Parameter_Wrap);
	lua_register(L, "Get_Int_Parameter", LuaScriptManager::Lua_Get_Int_Parameter_Wrap);
	lua_register(L, "Get_String_Parameter", LuaScriptManager::Lua_Get_String_Parameter_Wrap);
	lua_register(L, "Start_Timer", LuaScriptManager::Lua_Start_Timer_Wrap);
}
int LuaScriptManager::Lua_DestroyScript_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 2)
	{
		return 0;
	}
	int scriptID = (int)lua_tonumber(L, 1);
	int objID = (int)lua_tonumber(L, 2);
	LuaScriptManager::Lua_DestroyScript(Commands->Find_Object(objID), scriptID);
	return 0;
}
int LuaScriptManager::Lua_Get_Float_Parameter_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 3)
	{
		return 0;
	}
	int scriptID = (int)lua_tonumber(L, 1);
	int objID = (int)lua_tonumber(L, 2);
	const char *name = lua_tostring(L, 3);
	lua_pushnumber(L, LuaScriptManager::Lua_Get_Float_Parameter(Commands->Find_Object(objID), scriptID, name));
	return 1;
}
int LuaScriptManager::Lua_Get_Int_Parameter_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 3)
	{
		return 0;
	}
	int scriptID = (int)lua_tonumber(L, 1);
	int objID = (int)lua_tonumber(L, 2);
	const char *name = lua_tostring(L, 3);
	lua_pushnumber(L, LuaScriptManager::Lua_Get_Int_Parameter(Commands->Find_Object(objID), scriptID, name));
	return 1;
}
int LuaScriptManager::Lua_Get_String_Parameter_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 3)
	{
		return 0;
	}
	int scriptID = (int)lua_tonumber(L, 1);
	int objID = (int)lua_tonumber(L, 2);
	const char *name = lua_tostring(L, 3);
	lua_pushstring(L, LuaScriptManager::Lua_Get_String_Parameter(Commands->Find_Object(objID), scriptID, name));
	return 1;
}
int LuaScriptManager::Lua_Start_Timer_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 4)
	{
		return 0;
	}
	int scriptID = (int)lua_tonumber(L, 1);
	int obj = (int)lua_tonumber(L, 2);
	float Time = (float)lua_tonumber(L, 3);
	int Number = (int)lua_tonumber(L, 4);
	LuaScriptManager::Lua_Start_Timer(Commands->Find_Object(obj), scriptID, Time, Number);
	return 0;
}

//The dynamic script itself
void LuaScriptManager::LuaScript::Created(GameObject *obj)
{
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		if(strcmp((*Scripts.Node())->script, this->Get_Name()) == 0)
		{
			lua_State *L = (*Scripts.Node())->lua;
			lua_getglobal(L, "Created");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_call(L, 2, 0);
			}
		}
	}
}
void LuaScriptManager::LuaScript::Destroyed(GameObject *obj)
{
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		if(strcmp((*Scripts.Node())->script, this->Get_Name()) == 0)
		{
			lua_State *L = (*Scripts.Node())->lua;
			lua_getglobal(L, "Destroyed");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_call(L, 2, 0);
			}
		}
	}
}
void LuaScriptManager::LuaScript::Killed(GameObject *obj,GameObject *shooter)
{
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		if(strcmp((*Scripts.Node())->script, this->Get_Name()) == 0)
		{
			lua_State *L = (*Scripts.Node())->lua;
			lua_getglobal(L, "Killed");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, Commands->Get_ID(shooter));
				lua_call(L, 3, 0);
			}
		}
	}
}
void LuaScriptManager::LuaScript::Damaged(GameObject *obj,GameObject *damager,float damage)
{
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		if(strcmp((*Scripts.Node())->script, this->Get_Name()) == 0)
		{
			lua_State *L = (*Scripts.Node())->lua;
			lua_getglobal(L, "Damaged");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, Commands->Get_ID(damager));
				lua_pushnumber(L, damage);
				lua_call(L, 4, 0);
			}
		}
	}
}
void LuaScriptManager::LuaScript::Custom(GameObject *obj,int message,int param,GameObject *sender)
{
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		if(strcmp((*Scripts.Node())->script, this->Get_Name()) == 0)
		{
			lua_State *L = (*Scripts.Node())->lua;
			lua_getglobal(L, "Custom");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, message);
				lua_pushnumber(L, param);
				lua_pushnumber(L, Commands->Get_ID(sender));
				lua_call(L, 5, 0);
			}
		}
	}
}

void LuaScriptManager::LuaScript::Enemy_Seen(GameObject *obj,GameObject *seen)
{
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		if(strcmp((*Scripts.Node())->script, this->Get_Name()) == 0)
		{
			lua_State *L = (*Scripts.Node())->lua;
			lua_getglobal(L, "Enemy_Seen");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, Commands->Get_ID(seen));
				lua_call(L, 3, 0);
			}
		}
	}
}
void LuaScriptManager::LuaScript::Action_Complete(GameObject *obj,int action, ActionCompleteReason reason)
{
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		if(strcmp((*Scripts.Node())->script, this->Get_Name()) == 0)
		{
			lua_State *L = (*Scripts.Node())->lua;
			lua_getglobal(L, "Action_Complete");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, action);

				lua_call(L, 3, 0);
			}
		}
	}
}
void LuaScriptManager::LuaScript::Timer_Expired(GameObject *obj,int number)
{
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		if(strcmp((*Scripts.Node())->script, this->Get_Name()) == 0)
		{
			lua_State *L = (*Scripts.Node())->lua;
			lua_getglobal(L, "Timer_Expired");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, number);
				lua_call(L, 3, 0);
			}
		}
	}
}
void LuaScriptManager::LuaScript::Animation_Complete(GameObject *obj,const char *anim)
{
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		if(strcmp((*Scripts.Node())->script, this->Get_Name()) == 0)
		{
			lua_State *L = (*Scripts.Node())->lua;
			lua_getglobal(L, "Animation_Complete");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushstring(L, anim);
				lua_call(L, 3, 0);
			}
		}
	}
}
void LuaScriptManager::LuaScript::Poked(GameObject *obj,GameObject *poker)
{
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		if(strcmp((*Scripts.Node())->script, this->Get_Name()) == 0)
		{
			lua_State *L = (*Scripts.Node())->lua;
			lua_getglobal(L, "Poked");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, Commands->Get_ID(poker));
				lua_call(L, 3, 0);
			}
		}
	}
}
void LuaScriptManager::LuaScript::Entered(GameObject *obj,GameObject *enter)
{
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		if(strcmp((*Scripts.Node())->script, this->Get_Name()) == 0)
		{
			lua_State *L = (*Scripts.Node())->lua;
			lua_getglobal(L, "Entered");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, Commands->Get_ID(enter));
				lua_call(L, 3, 0);
			}
		}
	}
}
void LuaScriptManager::LuaScript::Exited(GameObject *obj,GameObject *exit)
{
	for(Scripts.ResetPos(); !Scripts.End(); Scripts.NodeNext())
	{
		if(strcmp((*Scripts.Node())->script, this->Get_Name()) == 0)
		{
			lua_State *L = (*Scripts.Node())->lua;
			lua_getglobal(L, "Exited");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, Commands->Get_ID(exit));
				lua_call(L, 3, 0);
			}
		}
	}
}



//LuaManager

AutoStack<lua_State *> LuaManager::Lua = AutoStack<lua_State *>();


void LuaManager::ShowBanner()
{
	printf("|---------------------------------------------------|\n");
	printf("| Lua 5.0 Copyright (C) 1994-2003 Tecgraf, PUC-Rio  |\n");
	printf("| R. Ierusalimschy, L. H. de Figueiredo & W. Celes  |\n");
	printf("| www.lua.org                                       |\n");
	printf("| Embedded with scripts.dll by RoShamBo             |\n");
	printf("| www.dcomproductions.net                           |\n");
	printf("|---------------------------------------------------|\n");
}

void LuaManager::LoadLua(const char *LuaFile)
{
	lua_State *L = lua_open();

	luaopen_io(L); 
	luaopen_base(L);
	luaopen_table(L);
	luaopen_string(L);
	luaopen_math(L);
	luaopen_loadlib(L);
	LuaSocketManager::Register(L);
	AddFunctions(L);
	int s = luaL_loadfile(L, LuaFile);
	if (s == 0) 
	{
		s = lua_pcall(L, 0, LUA_MULTRET, 0);
	}
	if(Report_Errors(L, s))
	{
		lua_close(L);
		printf("Failed to load plugin: %s\n", LuaFile);
		return;
	}
	printf("Loaded plugin: %s\n", LuaFile);
	Lua.Push(L);
}

void LuaManager::Cleanup()
{
	for(Lua.ResetPos(); !Lua.End(); Lua.NodeNext())
	{
		lua_close((*Lua.Node()));
	}
}

bool LuaManager::Report_Errors(lua_State *L, int status)
{
	if(status != 0) 
	{
		printf("%s\n", lua_tostring(L, -1));
		lua_pop(L, 1); 
		return 1;
	}
	return 0;
}

//hooks

void LuaManager::Call_Chat_Hook(int ID, int Type, const wchar_t *Msg)
{
	for(Lua.ResetPos(); !Lua.End(); Lua.NodeNext())
	{
		lua_State *L = (*Lua.Node());
		lua_getglobal(L, "OnChat");
		if(lua_isfunction(L, -1))
		{
			const char *Msg2 = WideCharToChar(Msg);
			lua_pushnumber(L, ID);
			lua_pushnumber(L, Type);
			lua_pushstring(L, Msg2);
			lua_call(L, 3, 0);
			delete []Msg2;
		}
	}
}
void LuaManager::Call_Host_Hook(int ID, int Type, const char *Msg)
{
	for(Lua.ResetPos(); !Lua.End(); Lua.NodeNext())
	{
		lua_State *L = (*Lua.Node());
		lua_getglobal(L, "OnHostMessage");
		if(lua_isfunction(L, -1))
		{
			lua_pushnumber(L, ID);
			lua_pushnumber(L, Type);
			lua_pushstring(L, Msg);
			lua_call(L, 3, 0);
		}
	}
}
void LuaManager::Call_Player_Join_Hook(int ID, const char *Nick)
{
	for(Lua.ResetPos(); !Lua.End(); Lua.NodeNext())
	{
		lua_State *L = (*Lua.Node());
		lua_getglobal(L, "OnPlayerJoin");
		if(lua_isfunction(L, -1))
		{
			lua_pushnumber(L, ID);
			lua_pushstring(L, Nick);
			lua_call(L, 2, 0);
		}
	}
}
void LuaManager::Call_Player_Leave_Hook(int ID)
{
	for(Lua.ResetPos(); !Lua.End(); Lua.NodeNext())
	{
		lua_State *L = (*Lua.Node());
		lua_getglobal(L, "OnPlayerLeave");
		if(lua_isfunction(L, -1))
		{
			lua_pushnumber(L, ID);
			lua_call(L, 1, 0);
		}
	}
}
void LuaManager::Call_Level_Loaded_Hook()
{
	for(Lua.ResetPos(); !Lua.End(); Lua.NodeNext())
	{
		lua_State *L = (*Lua.Node());
		lua_getglobal(L, "OnLevelLoaded");
		if(lua_isfunction(L, -1))
		{
			
			lua_call(L, 0, 0);
		}
	}
}
void LuaManager::Call_GameOver_Hook()
{
	for(Lua.ResetPos(); !Lua.End(); Lua.NodeNext())
	{
		lua_State *L = (*Lua.Node());
		lua_getglobal(L, "OnGameOver");
		if(lua_isfunction(L, -1))
		{
			lua_call(L, 0, 0);
		}
	}
}
void LuaManager::Call_Console_Output_Hook(const char *Output)
{
	for(Lua.ResetPos(); !Lua.End(); Lua.NodeNext())
	{
		lua_State *L = (*Lua.Node());
		lua_getglobal(L, "OnConsoleOutput");
		if(lua_isfunction(L, -1))
		{
			lua_pushstring(L, Output);
			lua_call(L, 1, 0);
		}
	}
}
void LuaManager::Call_Version_Hook(int ID, float Ver)
{
	for(Lua.ResetPos(); !Lua.End(); Lua.NodeNext())
	{
		lua_State *L = (*Lua.Node());
		lua_getglobal(L, "OnPlayerVersion");
		if(lua_isfunction(L, -1))
		{
			lua_pushnumber(L, ID);
			lua_pushnumber(L, Ver);
			lua_call(L, 2, 0);
		}
	}
}

void LuaManager::Purchase_Hook(BaseControllerClass *base, GameObject *purchaser, unsigned int cost, unsigned int preset,unsigned int purchaseret,const char *data)
{
	for(Lua.ResetPos(); !Lua.End(); Lua.NodeNext())
	{
		lua_State *L = (*Lua.Node());
		lua_getglobal(L, data);
		if(lua_isfunction(L, -1))
		{
			lua_pushnumber(L, Commands->Get_ID(purchaser));
			lua_pushnumber(L, cost);
			lua_pushnumber(L, preset);
			lua_pushnumber(L, purchaseret);
			lua_call(L, 1, 0);
		}
	}
}

void LuaManager::Call_Object_Hook(void *data, GameObject *obj)
{
	for(Lua.ResetPos(); !Lua.End(); Lua.NodeNext())
	{
		lua_State *L = (*Lua.Node());
		lua_getglobal(L, "OnObjectCreate");
		if(lua_isfunction(L, -1))
		{
			lua_pushnumber(L, Commands->Get_ID(obj));
			lua_call(L, 1, 0);
		}
	}
}

void LuaManager::Call_DDE_Hook(const char *DDEStr)
{
	for(Lua.ResetPos(); !Lua.End(); Lua.NodeNext())
	{
		lua_State *L = (*Lua.Node());
		lua_getglobal(L, "OnDDE");
		if(lua_isfunction(L, -1))
		{
			lua_pushstring(L, DDEStr);
			lua_call(L, 1, 0);
		}
	}
}




//socket manager

CRITICAL_SECTION LuaSocketManager::sync;
bool LuaSocketManager::init = 0;
AutoStack<LuaSocketManager::SocketInfo *> LuaSocketManager::Sockets = AutoStack<LuaSocketManager::SocketInfo *>();

void **LuaSocketManager::Connect(const char *Host, int Port)
{
	if(!init)
	{
		InitializeCriticalSection(&sync);
		init = 1;
	}
	void **ret = (void **)malloc(8);
	ret[0] = 0;
	ret[1] = 0;
	WSADATA wsainfo;
	if (WSAStartup(MAKEWORD(2,0), &wsainfo)) 
	{
		char *str = new char[32];
		sprintf(str, "Could not start Winsock2: %d", GetLastError());
		ret[1] = str;
		return ret;
    }
	hostent *he;
	if ((he = gethostbyname(Host)) == 0) 
	{
		char *str = new char[32];
		sprintf(str, "Could not resolve host: %d", GetLastError());
		ret[1] = str;
		return ret;
	}
	SOCKET s_ = socket(AF_INET,SOCK_STREAM,0);

	if (s_ == INVALID_SOCKET) 
	{
		char *str = new char[32];
		sprintf(str, "Could not create socket: %d", GetLastError());
		ret[1] = str;
		return ret;
	}
	sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_port = htons((u_short)Port);
	addr.sin_addr = *((in_addr *)he->h_addr);
	memset(&(addr.sin_zero), 0, 8); 
	if (connect(s_, (sockaddr *) &addr, sizeof(sockaddr))) 
	{
		char *str = new char[32];
		sprintf(str, "Could not connect: %d", GetLastError());
		ret[1] = str;
		return ret;
	}
	int *ID = new int;
	(*ID) = 0;
	SocketInfo *SockIfo = new SocketInfo;
	SockIfo->Socket = new SOCKET;
	*SockIfo->Socket = s_;

	for(Sockets.ResetPos(); !Sockets.End(); Sockets.NodeNext())
	{
		(*ID)++;
	}
	SockIfo->ID = *ID;
	Sockets.Push(SockIfo);
	SockIfo->Thread = CreateThread(NULL, NULL, Thr, SockIfo, NULL, NULL);

	ret[0] = ID;
	return ret;
	
}


void LuaSocketManager::Send(int ID, const char *Data)
{
	if(!init)
	{
		InitializeCriticalSection(&sync);
		init = 1;
	}
	EnterCriticalSection(&sync);
	for(Sockets.ResetPos(); !Sockets.End(); Sockets.NodeNext())
	{
		SocketInfo *sockifo = (*Sockets.Node());
		if(sockifo->ID == ID)
		{
			send(*sockifo->Socket, Data, (int)strlen(Data), 0);
		}
	}
	LeaveCriticalSection(&sync);
}

const char *LuaSocketManager::Recv(int ID)
{
	if(!init)
	{
		InitializeCriticalSection(&sync);
		init = 1;
	}
	EnterCriticalSection(&sync);
	for(Sockets.ResetPos(); !Sockets.End(); Sockets.NodeNext())
	{
		SocketInfo *sockifo = (*Sockets.Node());
		if(sockifo->ID == ID)
		{
			DWORD code = 1;
			if(GetExitCodeThread(sockifo->Thread, &code))
			{
				if(code == 0)
				{
					CloseHandle(sockifo->Thread);
					
					sockifo->buffer.DeleteAllNodes(1);
					delete sockifo->Socket;
					AutoStack<SocketInfo *> tmp;
					for(Sockets.ResetPos(); !Sockets.End(); Sockets.NodeNext())
					{
						if((*Sockets.Node())->ID != sockifo->ID)
						{
							tmp.Push((*Sockets.Node()));
						}
					}
					Sockets = tmp;
					delete sockifo;
					LeaveCriticalSection(&sync);
					char *ret = new char[2];
					ret[0] = (char)4;
					ret[1] = '\0';
					return ret;
				}
			}



			sockifo->buffer.ResetPos();
			if(sockifo->buffer.End())
			{
				LeaveCriticalSection(&sync);
				return 0;
			}
			else
			{
				const char *ret = (*sockifo->buffer.Node());
				AutoStack<char *> tmp;

				for(sockifo->buffer.NodeNext();!sockifo->buffer.End(); sockifo->buffer.NodeNext())
				{
					tmp.Push((*sockifo->buffer.Node()));
				}
				sockifo->buffer = tmp;
				LeaveCriticalSection(&sync);
				return ret;
			}
		}
	}
	LeaveCriticalSection(&sync);
	return 0;
}

DWORD WINAPI LuaSocketManager::Thr(LPVOID param)
{
	SocketInfo *ifo = (SocketInfo *)param;
	SOCKET sock = (*ifo->Socket);
	
	while(ifo)
	{
		char buffer[4096];
		int i = recv(sock, buffer, 4095, 0);
		
		if(i <= 0)
		{
			break;
		}
		buffer[i] = '\0';
		char *data = new char[i+1];
		strcpy(data, buffer);

		EnterCriticalSection(&sync);
		
		for(Sockets.ResetPos(); !Sockets.End(); Sockets.NodeNext())
		{
			SocketInfo *sockifo = (*Sockets.Node());
			if(sockifo->ID == ifo->ID)
			{
				sockifo->buffer.Push(data);
			}
		}

		LeaveCriticalSection(&sync);
	
	}
	return 0;
}

void LuaSocketManager::Disconnect(int ID)
{
	if(!init)
	{
		InitializeCriticalSection(&sync);
		init = 1;
	}
	EnterCriticalSection(&sync);
	AutoStack<SocketInfo *> tmp;
	SocketInfo *close = 0;
	for(Sockets.ResetPos(); !Sockets.End(); Sockets.NodeNext())
	{
		SocketInfo *sockifo = (*Sockets.Node());
		if(sockifo->ID != ID)
		{
			tmp.Push(sockifo);
		}
		else
		{
			close = sockifo;
		}
	}
	if(!close)
	{
		LeaveCriticalSection(&sync);
		return;
	}
	else
	{
		Sockets = tmp;
		send(*close->Socket, "", 0, 0);
		closesocket(*close->Socket);
		delete close->Socket;
		close->buffer.DeleteAllNodes(1);
		delete close;
	}


	LeaveCriticalSection(&sync);
}


int LuaSocketManager::Lua_Connect_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 2)
	{
		return 0;
	}
	const char *Host = lua_tostring(L, 1);
	int Port = (int)lua_tonumber(L, 2);
	void **ret = LuaSocketManager::Connect(Host, Port);
	if(ret[0] == 0)
	{
		lua_pushstring(L, (char *)ret[1]);
		delete [](char *)ret[1];
		free(ret);
		return 1;
		
	}
	else
	{
		int ID = (*(int *)ret[0]);
		lua_pushnumber(L, ID);
		delete (int *)ret[0];
		free(ret);
		return 1;
	}
	
}
int LuaSocketManager::Lua_Send_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 2)
	{
		return 0;
	}
	int ID = (int)lua_tonumber(L, 1);
	const char *data = lua_tostring(L, 2);
	LuaSocketManager::Send(ID, data);
	return 0;
}
int LuaSocketManager::Lua_Recv_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 1)
	{
		return 0;
	}
	int ID = (int)lua_tonumber(L, 1);
	const char *data = LuaSocketManager::Recv(ID);
	if(!data)
	{
		return 0;
	}
	else
	{
		lua_pushstring(L, data);
		delete []data;
		return 1;
	}
}

bool LuaSocketManager::ValidSock(int ID)
{
	if(!init)
	{
		InitializeCriticalSection(&sync);
		init = 1;
	}
	EnterCriticalSection(&sync);
	for(Sockets.ResetPos(); !Sockets.End(); Sockets.NodeNext())
	{
		if((*Sockets.Node())->ID == ID)
		{
			LeaveCriticalSection(&sync);
			return 1;
		}
	}
	LeaveCriticalSection(&sync);
	return 0;
}

int LuaSocketManager::Lua_ValidSock_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 1)
	{
		return 0;
	}
	int ID = (int)lua_tonumber(L, 1);
	bool ret = ValidSock(ID);
	if(ret)
	{
		lua_pushnumber(L, 1);
	}
	else
	{
		lua_pushnumber(L, 0);
	}
	return 1;
}
void LuaSocketManager::Register(lua_State *L)
{
	lua_register(L, "Connect", Lua_Connect_Wrap);
	lua_register(L, "Send", Lua_Send_Wrap);
	lua_register(L, "Recv", Lua_Recv_Wrap);
	lua_register(L, "ValidSock", Lua_ValidSock_Wrap);
	lua_register(L, "Disconnect", Lua_Disconnect_Wrap);
}

void LuaSocketManager::Cleanup()
{
	if(!init)
	{
		InitializeCriticalSection(&sync);
		init = 1;
	}
	EnterCriticalSection(&sync);
	for(Sockets.ResetPos(); !Sockets.End(); Sockets.NodeNext())
	{
		DWORD code = 1;
		if(GetExitCodeThread((*Sockets.Node())->Thread, &code))
		{
			if(code != 0)
			{
				TerminateThread((*Sockets.Node())->Thread, 0);
			}
		}
		CloseHandle((*Sockets.Node())->Thread);
		delete (*Sockets.Node())->Socket;
		(*Sockets.Node())->buffer.DeleteAllNodes(1);
		delete (*Sockets.Node());
	}
	Sockets.RemoveAllNodes();
	LeaveCriticalSection(&sync);
}

int LuaSocketManager::Lua_Disconnect_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 1)
	{
		return 0;
	}
	int ID = (int)lua_tonumber(L, 1);
	Disconnect(ID);
	return 0;
}