#include "scripts.h"
#include "engine.h"
#include "stdlib.h"
#include "stdio.h"
#include <Windows.h>
#include "Misc.h"
//#define __MEMDEBUG


AutoStack<meminfo> alloc;

void* operator new(size_t nbytes, const char *File, int Line)
{
	void *ptr = AllocMemory(nbytes);

	meminfo ifo;
	ifo.File = File;
	ifo.Line = Line;
	ifo.Address = ptr;
	alloc.Push(ifo);
#ifdef __MEMDEBUG
	printf("Allocated memory address %p (%d bytes) in %s:%d\n", ptr, nbytes, File, Line);
#endif
	return ptr;
}

void* operator new[](size_t nbytes, const char *File, int Line)
{
	return operator new(nbytes, File, Line);
}
void* operator new(size_t nbytes)
{
	void *ptr = AllocMemory(nbytes);

	meminfo ifo;
	ifo.File = "Unknown";
	ifo.Line = 0;
	ifo.Address = ptr;
	alloc.Push(ifo);
#ifdef __MEMDEBUG
	printf("Allocated memory address %p (%d bytes)\n", ptr, nbytes);
#endif
	return ptr;
}

void* operator new[](size_t nbytes)
{
	return operator new(nbytes);
}

void operator delete(void *ptr)
{
	AutoStack<meminfo> tmp;
	meminfo ifo = meminfo();
	ifo.File = 0;
	for(alloc.ResetPos(); !alloc.End(); alloc.NodeNext())
	{
		if((*alloc.Node()).Address != ptr)
		{
			tmp.Push((*alloc.Node()));
		}
		else
		{
			ifo = (*alloc.Node());
		}
	}
	alloc = tmp;
#ifdef __MEMDEBUG
	printf("Deallocated memory address %p at %s:%d\n", ptr, ifo.File, ifo.Line);
#endif
	if(!ifo.File)
	{
		free(ptr);
	}
	else
	{
		DeallocMemory(ptr);
	}
}

void operator delete[](void *ptr)
{
	operator delete(ptr);
}



void check_memory()
{
	alloc.ResetPos();
	if(!alloc.End())
	{
		HANDLE file = (HANDLE)fopen("C:\test.txt", "w");
		if(!file)
		{
			for(;!alloc.End(); alloc.NodeNext())
			{
				printf("unfreed memory! %s:%d\n", (*alloc.Node()).File, (*alloc.Node()).Line);
			}
			return;
		}
		char buf[2048];
		buf[0] = '\0';
		for(;!alloc.End(); alloc.NodeNext())
		{
			sprintf(buf, "%sunfreed memory! %s:%d\n", buf, (*alloc.Node()).File, (*alloc.Node()).Line);
		}
		fputs(buf, (FILE *)file);
		fclose((FILE *)file);
	}
}

AutoStack<char *> GetFiles(const char *Path)
{
	AutoStack<char *> Dirs;
	WIN32_FIND_DATA FindFileData;
	HANDLE hFind = INVALID_HANDLE_VALUE;

	hFind = FindFirstFile(Path, &FindFileData);
	if(FindFileData.dwFileAttributes != FILE_ATTRIBUTE_DIRECTORY)
	{
		char *file = new char[strlen(FindFileData.cFileName)+1];
		strcpy(file, FindFileData.cFileName);
		Dirs.Push(file);
	}
	while (FindNextFile(hFind, &FindFileData) != 0) 
	{
		if(FindFileData.dwFileAttributes != FILE_ATTRIBUTE_DIRECTORY)
		{
			char *file = new char[strlen(FindFileData.cFileName)+1];
			strcpy(file, FindFileData.cFileName);
			Dirs.Push(file);
		}
	}
	FindClose(hFind);
	return Dirs;

}